// romto8xu.cpp: dfinit le point d'entre pour l'application console.
//
#define CKOFFSET2	0x24FFF8 // fx-9750/9860G OS checksum
#define CKOFFSET1	0x1DFFF8 // fx-7400Gii OS checksum
#define ROM_SIZE 2555904

#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "bmp.h"
#include "g2_offbmp_injector.h"

#define DEPTH	1
#define FILENAMAXSIZE 30

typedef unsigned char byte;

void printinfos()
{	printf("\n+----------------------------------+\n");
	printf("!         G2_BMP_INJECTOR_PRO      !\n");
	printf("+----------------------------------+\n");
	printf(" X. Andreani & L. Gatin\n\n");
	printf("Injects a BMP off-image or main menu banner\n");
	printf("in a Casio graphic calculator OS.\n");
	printf("BMP image has to be monochrome (1-bit depth)\n");
    printf("off-image : 128*64, banner : 128*7\n");
	printf("Supported models:\n");
	printf("* Graph 25+Pro / fx-7400Gii\n");
	printf("* Graph 35+USB / fx-9750Gii\n");
	printf("* Graph 85(SD) / fx-9860G(SD)\n");
	printf("* Graph 75/95 / fx-9860Gii(SD)\n\n");
}

void printerror(char* msg){
	printf("ERROR: %s\n",msg);
}

byte *bmpdata = NULL;

int main(int argc, char *argv[])
{
    byte* buffer;
    int WIDTH;
    int HEIGHT;
    int tmp;
    int i;

	int romsize=ROM_SIZE;
	char type;
	char *arg1 = NULL;
	char *arg2 = NULL;
	char *arg3 = NULL;

    byte oldlogo[8];

	FILE* rom;
	FILE* polyrom;
	FILE* bmp;

    arg1 = (char*)malloc(FILENAMAXSIZE * sizeof(char));
    arg2 = (char*)malloc(FILENAMAXSIZE * sizeof(char));
    arg3 = (char*)malloc(FILENAMAXSIZE * sizeof(char));

    printinfos();

	printf("Please enter name of OS source file\n");
	scanf("%s", arg1);
    printf("Please enter name of BMP image\n");
	scanf("%s", arg2);
	printf("Please enter name of outgoing OS\n");
	scanf("%s", arg3);

	while(1)
    {
        type = -1;
        printf("Please enter type of injection\n");
        printf("off-image: 0, banner: 1\n");
        scanf("%d", &type);

        if(type==0 || type==1) break;
        else printf("Must be 0 or 1 !\n");
    }

    if(type==0)
    {
        byte tmp2[8] = {0,0,0,0x01,0xFF,0xC0,0x3E,0x03};
        for(i=0; i<8; ++i) oldlogo[i] = tmp2[i];
        WIDTH = 128;
        HEIGHT = 64;
        tmp = -0x1A0;
    }
    else
    {
        byte tmp2[8] = {0x92, 0x49, 0x24, 0x92, 0x42, 0x27, 0x1C, 0x88};
        for(i=0; i<8; ++i) oldlogo[i] = tmp2[i];
        WIDTH = 128;
        HEIGHT = 7;
        tmp = 0;
    }

    bmpdata = (byte*)malloc((WIDTH*HEIGHT*DEPTH/8)*sizeof(char));

    rom=fopen(arg1,"rb");
		if(!rom)
			printerror("Source ROM file open error");
		else
		{	fseek(rom,0,SEEK_END);
			romsize=ftell(rom);
			fseek(rom,0,SEEK_SET);
			if(romsize>ROM_SIZE) romsize=ROM_SIZE;
			printf("ROM file size is %i.\n",romsize);
			if(romsize<ROM_SIZE)	printf("Assuming fx-7400Gii ROM file.\n");
			else					printf("Assuming fx-9750/9860G ROM file.\n");
			buffer = (byte*) malloc(romsize*sizeof(byte));
			memset(buffer,0xFF,romsize);
			printf("Buffering ROM file data...\n");
			int size=0;
			while(size<romsize)
				size+=fread(buffer+size,sizeof(byte),romsize-size,rom);
			fclose(rom);
			printf("Looking for off-logo data in the OS...\n");
			int found=0;
			int i;
			for(i=0x10000; i<size && !found; i++)
			{	if(!memcmp(oldlogo,buffer+i,8))
					found=1;
			}
			i--;
			if(found==0)
				printerror("No such data found");
			else
			{	byte* ptr=buffer+i+tmp;
				printf("Off-logo data found at offset 0x%x!\n",ptr-buffer);
				bmp=fopen(arg2,"rb");
				if(!bmp)
					printerror("BMP file open error\n");
				else
				{	fileheader bmpfheader;
					if(!readFileHeader(&bmpfheader,bmp))
						printerror("BMP file header read error");
					else
					{	if(!bmpfheader.type)
							printerror("Not a BMP file");
						else
						{	dataheader bmpdheader;
							if(!readDataHeader(&bmpdheader,bmp,&bmpfheader))
								printerror("BMP data header read error");
							else
							{	printf("BMP %i x %i x %i-bit(s) ",bmpdheader.width,bmpdheader.height,bmpdheader.bits);
								if(bmpdheader.compression)	printf("compressed\n");
								else						printf("RAW\n");
								if(bmpdheader.width!=WIDTH || bmpdheader.height!=HEIGHT || bmpdheader.bits!=1)
								{	printerror("Wrong BMP format");
									printf("We want %i x %i x %i-bit(monochrome) RAW data\n",WIDTH,HEIGHT,DEPTH);
								}
								else
								{	int* palette=0;
									palette=(int*)malloc(bmpdheader.indexedcolors*sizeof(int));
									if(!readPalette(palette, &bmpdheader, bmp, &bmpfheader))
										printerror("BMP palette read error");
									else
									{	int i;
										if(palette[0]==palette[1])
										{	palette[0]=0;
											palette[1]=0xFFFFFF;
										}
										if(palette[0]>palette[1])
										{	palette[0]=0; // black LCD
											palette[1]=1; // white LCD
										}
										else
										{	palette[0]=1; // black LCD
											palette[1]=0; // white LCD
										}
										if(!readImage(bmpdata, &bmpdheader, palette, bmp, &bmpfheader))
											printerror("Read image data");
										else
										{	memcpy(ptr,bmpdata,WIDTH*HEIGHT*DEPTH/8);
											int checksum=0;
											int ckoffset=CKOFFSET2;
											if(romsize<ROM_SIZE) ckoffset=CKOFFSET1;
											for(i=0x10000;i<ckoffset;i++)
											{	checksum+=buffer[i];
												if(checksum>0x100000000) checksum-=0x100000000;
											}
											printf("New checksum is 0x%x.\n",checksum);
											for(i=3;i>=0;i--)
											{	buffer[i+ckoffset]=checksum%256;
												checksum=checksum/256;
											}
											polyrom=fopen(arg3,"wb");
											if(!polyrom)
												printerror("Dest ROM file open error");
											else
											{	printf("Flushing ROM file data...\n");
												size=0;
												while(size<romsize)
													size+=fwrite(buffer+size,sizeof(byte),romsize-size,polyrom);
												fclose(polyrom);
												printf("DONE !\n");
												system("pause");
											}
										}
									}
								}
							}
						}
					}
					fclose(bmp);
				}
			}
			free(buffer);
			fclose(rom);
		}
	return 0;
}
